/*
 * Copyright (c) 2017-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * See file LICENSE for terms.
 */

#ifndef THREAD_POOL_H
#define THREAD_POOL_H

#include <pthread.h>
#include <list>
#include <queue>

#include "agg_types.h"

#define DEFAULT_THREADPOOL_THREADS_NUMBER 5

typedef std::queue<class ThreadPoolTask*> TaskQueue;
typedef std::list<pthread_t> ListThreads;

class ThreadPoolTask
{
    bool m_check_result_;

   public:
    // Run should not throw for error handling.
    // It should change the tasks state instead.
    virtual void Run() = 0;
    virtual ~ThreadPoolTask() {}

    void SetCheckResult(bool is_check_result) { m_check_result_ = is_check_result; }

    bool IsCheckResult() { return m_check_result_; }

    virtual void Finalize(){};
};

class ThreadPool
{
    TaskQueue m_task_queue_;
    ListThreads m_threads_;
    bool m_stop_;
    bool m_init_;

    pthread_mutex_t m_queue_lock_;
    pthread_cond_t m_queue_cond_;

   public:
    ThreadPool() : m_stop_(false), m_init_(false) {}

    int Init(uint16_t num_threads);

    void Stop() { m_stop_ = true; }
    void AddTask(ThreadPoolTask* p_task);

    // the method executed by each thread
    void ThreadRun();

    ~ThreadPool();
};

class ThreadPoolTasksCollection
{
   private:
    uint16_t m_num_tasks_in_progress_;
    pthread_mutex_t m_tasks_lock_;
    pthread_cond_t m_tasks_cond_;

   protected:
    bool m_is_init_;

   public:
    ThreadPoolTasksCollection() : m_num_tasks_in_progress_(0), m_is_init_(false) {}

    ~ThreadPoolTasksCollection();

    void Init();

    void AddTaskToThreadPool(ThreadPool& thread_pool, ThreadPoolTask* p_task);

    void WaitForTasks();

    // this should be called before each tasks ends
    void OnTaskEnd();
};

#endif   // THREAD_POOL_H
